"""
"""


#[
from __future__ import annotations

from collections.abc import (Iterable, )
from typing import (Self, Any, NoReturn, TypeAlias, Protocol, )
import re as _re
import parsimonious as _pa
import functools as _ft
import itertools as _it
import copy as _cp

from .. import wrongdoings as _wrongdoings
from . import _pseudofunctions as _pseudofunctions
from . import common as _common
from . import _shifts as _shifts
from . import _lists as _lists
#]


def from_string(
    source: str,
    *,
    context: dict[str, Any] | None = None,
    save_preparsed: str = "",
) -> tuple[str, dict]:
    """
    """
    #[
    info = {
        "context": None,
        "preparser_needed": None,
    }

    context = dict(context) if context else {}
    info["context"] = context

    # Remove NON-NESTED block comments #{ ... #} or %{ ... %}
    source = _remove_block_comments(source, )

    # Remove line comments %, #, ..., \, except for #!, %!
    source = _remove_line_comments(source, )

    # Replace time shifts {...} by [...]
    source = _shifts.standardize_time_shifts(source)

    # Replace human prefix (!) with processing prefix (¡)
    source = _translate_keywords(source)

    # Add blank lines pre and post source
    source = _common.add_blank_lines(source)

    # Run preparser on !-commands only if necessary
    preparser_needed = _is_preparser_needed(source)
    preparsed_source = (
        _run_preparser_on_source_string(source, context, )
        if preparser_needed else source
    )

    # Remove multiple blank lines
    preparsed_source = _consolidate_lines(preparsed_source, )

    # Evaluate and stringify Python expressions in <...> or <<...>>
    # Do this only after the !-command preparser because the <...> expression
    # may be ?-dependent
    preparsed_source = _evaluate_contextual_expressions(preparsed_source, info["context"], )

    # Expand lists
    preparsed_source = _lists.resolve_lists(preparsed_source, )

    # Resolve pseudofunctions: diff(...), ...
    preparsed_source = _pseudofunctions.resolve_pseudofunctions(preparsed_source, )

    if save_preparsed:
        _save_preparsed_source(preparsed_source, save_preparsed, )

    info = {
        "preparser_needed": preparser_needed,
        "preparsed_source": preparsed_source,
    }

    return preparsed_source, info
    #]


_GRAMMAR_DEF = _common.GRAMMAR_DEF + r"""

    source =  (control / text / list)+
    control = white_spaces (for_do_block / if_then_block / else_keyword / end_keyword)
    text = ~r"[^¡]+"

    keyword = (
        for_keyword / do_keyword / 
        if_keyword / then_keyword / else_keyword /
        end_keyword /
        list_keyword
    )

    for_do_block = for_keyword white_spaces for_control do_keyword
        for_control = for_control_name_equals? for_tokens
        for_control_name_equals = for_control_name white_spaces ~r"[=:]" white_spaces
        for_control_name = ~r"\?[\w\(\)]*"
        for_tokens = ~r"\s*[^¡]+\s*"
        for_keyword = keyword_prefix "for"
        do_keyword = keyword_prefix "do"

    if_then_block = if_keyword if_condition then_keyword
        if_condition = ~r"[^¡]+"
        if_keyword = keyword_prefix "if"
        then_keyword = keyword_prefix "then"

    else_keyword = keyword_prefix "else"
    end_keyword = keyword_prefix "end"

    list = list_keyword "(" list_name ")"
    list_keyword = keyword_prefix "list"
    list_name = ~r"`\w+"

"""


_GRAMMAR = _pa.grammar.Grammar(_GRAMMAR_DEF)
_KEYWORDS = [ k.members[1].literal for k in _GRAMMAR["keyword"].members ]
_KEYWORDS_PATTERN = _common.compile_keywords_pattern(_KEYWORDS, )
_translate_keywords = _ft.partial(_common.translate_keywords, _KEYWORDS_PATTERN, )


class _Visitor(_pa.nodes.NodeVisitor):
    """
    """
    #[
    def __init__(self, context=None):
        super().__init__()
        self.content = []
        self.context = dict(context) if context else {}

    def _add(self, new):
        self.content.append(new)

    def visit_source(self, node, visited_children, ):
        return self.content

    def visit_end_keyword(self, node, visited_children, ):
        self._add(_End())

    def visit_else_keyword(self, node, visited_children, ):
        self._add(_Else())

    def visit_for_do_block(self, node, visited_children, ):
        control_name, tokens_as_string = visited_children[2]
        self._add(_For(control_name, tokens_as_string))

    def visit_for_control(self, node, visited_children, ):
        control_name = visited_children[0][0] if visited_children[0] else "?"
        tokens = visited_children[1]
        return [control_name, tokens]

    def visit_for_tokens(self, node, visited_children, ):
        return node.text

    def visit_for_control_name_equals(self, node, visited_children, ):
        return visited_children[0]

    def visit_if_then_block(self, node, visited_children, ):
        condition = visited_children[1].strip()
        self._add(_If(condition))

    def visit_text(self, node, visited_children, ):
        text = node.text
        if text:
            self._add(_Text(text))

    def visit_list(self, node, visited_children, ):
        self._add(_Text(node.text))

    def generic_visit(self, node, visited_children, ):
        return visited_children or node.text
    #]


class _DirectiveProtocol(Protocol, ):
    def resolve(self, sequence, context, ): ...
    def replace(self, pattern, replacement, ): ...


Sequence: TypeAlias = Iterable[_DirectiveProtocol]


class _Text:
    """
    """
    #[

    level = 0

    def __init__(self, content: str, ):
        self.content = content

    def replace(self, pattern: str, replacement: str, ):
        new = _Text(self.content.replace(pattern, replacement))
        return new

    def resolve(self, sequence: Sequence, context: dict, ):
        return self.content, sequence[1:]
    #]


class _For:
    """
    """
    #[

    level = 1

    def __init__(self, control_name, tokens_as_string, /, ):
        self._control_name = control_name
        self._tokens_as_string = tokens_as_string

    def replace(self, pattern, replacement, /, ) -> Self:
        self._tokens_as_string = self._tokens_as_string.replace(pattern, replacement, )
        return self

    def _expand_tokens(self, for_body_sequence: Sequence, context: dict, /, ):
        """
        """
        tokens = self._prepare_tokens(context, )
        control_name = self._control_name
        bare_control_name = control_name.removeprefix("?")
        #
        if bare_control_name.startswith("(") and bare_control_name.endswith(")"):
            upper_control_name = control_name.replace("(", "{", ).replace(")", "}", )
            lower_control_name = control_name.replace("(", "[", ).replace(")", "]", )
            control_name_upper = control_name + "|upper"
            control_name_lower = control_name + "|lower"
            def replace_control_by_token(source_text, token):
                token_upper = token.upper()
                token_lower = token.lower()
                source_text = source_text.replace(upper_control_name, token_upper, )
                source_text = source_text.replace(lower_control_name, token_lower, )
                source_text = source_text.replace(control_name_upper, token_upper, )
                source_text = source_text.replace(control_name_lower, token_lower, )
                source_text = source_text.replace(control_name, token, )
                return source_text
        else:
            def replace_control_by_token(source_text, token):
                source_text = source_text.replace(control_name, token, )
                return source_text
        #
        new = []
        for t in tokens:
            new.extend(
                replace_control_by_token(s, t, )
                for s in _cp.deepcopy(for_body_sequence, )
            )
        return new

    def _prepare_tokens(self, context: dict, ):
        """
        """
        tokens_as_string = _evaluate_contextual_expressions(self._tokens_as_string, context, )
        return _re.findall(r"\w+", tokens_as_string, )

    def resolve(self, sequence: Sequence, context: dict, /, ):
        """
        """
        index_end = _find_matching_end(sequence)
        for_body_sequence = self._expand_tokens(sequence[1:index_end], context, )
        code = _resolve_sequence(for_body_sequence, context, )
        return code, sequence[index_end+1:]

    #]


class _If:
    """
    """
    #[
    level = 1

    def __init__(self, condition, /, ):
        self._condition_text = condition.strip()
        self._condition_result = None

    def replace(self, pattern, replacement) -> Self:
        self._condition_text = self._condition_text.replace(pattern, replacement)
        return self

    def resolve(self, sequence: Sequence, context: dict, /, ):
        """
        """
        self._evaluate_condition(context, )
        index_end = _find_matching_end(sequence, )
        index_else = _find_matching_else(sequence, ) or index_end
        if self._condition_result:
            code = _resolve_sequence(sequence[1:index_else], context, )
        else:
            code = _resolve_sequence(sequence[index_else+1:index_end], context, )
        return code, sequence[index_end+1:]

    def _evaluate_condition(
        self,
        context: dict[str, Any] | None,
        /,
    ) -> None:
        """
        """
        context = dict(context) if context else {}
        try:
            self._condition_result = bool(eval(self._condition_text, {}, context, ))
        except Exception as exc:
            raise _wrongdoings.IrisPieError(
                f"Failed to evaluate this !if condition: {self._condition_text}",
            ) from exc
    #]


class _Else:
    """
    """
    #[
    level = 0

    def replace(self, pattern, replacement):
        """
        """
        return self

    def resolve(self, *args, **kwargs, ) -> NoReturn:
        """
        """
        raise _wrongdoings.IrisPieError(
            "Misplaced preparsing directive !else",
        )
    #]


class _End:
    """
    """
    #[

    level = -1

    def replace(self, pattern, replacement):
        return self

    def resolve(self, *args, **kwargs, ) -> NoReturn:
        """
        """
        raise _wrongdoings.IrisPieError(
            "Misplaced preparsing directive !end",
        )

    #]


_LINE_COMMENT_PATTERN = _re.compile(r'"[^"\n]*"|[%#](?!!).*|\.\.\..*|\\.*', )
_BLOCK_COMMENT_PATTERN = _re.compile(r"([%#]){.*?\1\}", _re.DOTALL, )


def _remove_line_comments(source: str, /, ) -> str:
    return _re.sub(
        _LINE_COMMENT_PATTERN,
        lambda m: m.group(0) if m.group(0).startswith('"') else "",
        source,
    )

def _remove_block_comments(source: str, /, ) -> str:
    return _re.sub(_BLOCK_COMMENT_PATTERN, "", source, )

def _cumulate_level(sequence: Sequence, /, ) -> int:
    return list(_it.accumulate(s.level for s in sequence))


def _find_matching_end(sequence: Sequence, /, ) -> int:
    """
    Find !end that is on level 0
    """
    #[
    cum_level = _cumulate_level(sequence, )
    return cum_level.index(0)
    #]


def _find_matching_else(sequence: Sequence, /, ) -> int:
    """
    Find !else that is on level 0
    """
    #[
    cum_level = _cumulate_level(sequence, )
    return next(
        (i for i, l in enumerate(cum_level) if l == 1 and isinstance(sequence[i], _Else)),
        None,
    )
    #]


def _resolve_sequence(sequence: Sequence, context: dict, /, ) -> str:
    """
    """
    #[
    code = ""
    while sequence:
        new_code, sequence = sequence[0].resolve(sequence, context, )
        if code and new_code:
            code = code + "\n" + new_code
        elif new_code:
            code = new_code
        else:
            pass
    return code
    #]


def _consolidate_lines(text: str) -> str:
    split_text = _re.split(" *\n+", text, )
    return "\n".join(i for i in split_text if i)


_CONTEXTUAL_EXPRESSION_PATTERN = _re.compile(r"<+([^>]*)>+")


def _evaluate_contextual_expressions(text: str, context: dict, /):
    """
    """
    #[
    def _replace(match):
        try:
            expression = match.group(1).strip()
            return _stringify(eval(expression, {}, context, ), )
        except:
            raise Exception(f"Failed to evaluate this contextual expression: {expression}")
    return _re.sub(_CONTEXTUAL_EXPRESSION_PATTERN, _replace, text)
    #]


def _stringify(input, /, ):
    if isinstance(input, str):
        return input
    elif not isinstance(input, Iterable):
        return str(input)
    else:
        return ",".join(_stringify(i, ) for i in input)


def _is_preparser_needed(source: str, /, ) -> bool:
    return _common.KEYWORD_PREFIX in source


def _save_preparsed_source(source: str, file_name: str, /, ) -> None:
    if file_name:
        with open(file_name, "wt+") as fid:
            fid.write(source)

def _run_preparser_on_source_string(source: str, context: dir, /, ) -> str:
    nodes = _GRAMMAR["source"].parse(source, )
    visitor = _Visitor()
    sequence = visitor.visit(nodes, )
    return _resolve_sequence(sequence, context, )





"""
Evaluators for dynamic period-by-period systems
"""


#[
from __future__ import annotations

import numpy as _np

from .. import equations as _equations
from .. import quantities as _quantities

from . import printers as _printers
from . import equators as _equators
from . import jacobians as _jacobians

from typing import (TYPE_CHECKING, )
if TYPE_CHECKING:
    from typing import (Any, Callable, )
    from collections.abc import (Iterable, )
#]


class PeriodEvaluator:
    """
    """
    #[

    def __init__(
        self,
        wrt_qids: list[int],
        wrt_equations: Iterable[_equations.Equation],
        all_quantities: Iterable[_quantities.Quantity],
        *,
        terminal_simulator: Callable | None = None,
        context: dict | None = None,
        iter_printer_settings: dict[str, Any] | None = None,
    ) -> None:
        """
        """
        self._wrt_qids = list(wrt_qids, )
        self.terminal_simulator = terminal_simulator
        qid_to_logly = _quantities.create_qid_to_logly(all_quantities, )
        qid_to_name = _quantities.create_qid_to_name(all_quantities, )
        #
        # Index of loglies within wrt_qids; needs to be list not tuple
        # because of numpy indexing
        self._index_logly = \
            list(_quantities.generate_where_logly(self._wrt_qids, qid_to_logly, ))
        #
        # Set up components
        self._equator = _equators.PeriodEquator(
            wrt_equations,
            context=context,
        )
        self._jacobian = _jacobians.PeriodJacobian(
            wrt_equations,
            self._wrt_qids,
            qid_to_logly,
            context=context,
        )
        self.iter_printer = _printers.PeriodIterPrinter(
            wrt_equations,
            self._wrt_qids,
            qid_to_logly,
            qid_to_name,
            **(iter_printer_settings or {}),
        )

    def get_init_guess(
        self,
        data_array: _np.ndarray,
        column: int,
    ) -> _np.ndarray:
        """
        """
        maybelog_guess = data_array[self._wrt_qids, column]
        maybelog_guess[self._index_logly] = _np.log(maybelog_guess[self._index_logly])
        return maybelog_guess

    def update(
        self,
        maybelog_guess: _np.ndarray,
        data_array: _np.ndarray,
        column: int,
    ) -> None:
        """
        """
        #
        # Create a copy of maybelog_guess to avoid modifying the original
        # array in scipy.optimize.root
        maybelog_guess = _np.copy(maybelog_guess, )
        maybelog_guess[self._index_logly] = \
            _np.exp(maybelog_guess[self._index_logly])
        data_array[self._wrt_qids, column] = maybelog_guess

    def eval(
        self,
        maybelog_guess: _np.ndarray,
        data_array: _np.ndarray,
        column: int,
    ) -> _np.ndarray:
        """
        """
        self.update(maybelog_guess, data_array, column, )
        if self.terminal_simulator:
            self.terminal_simulator(data_array, column, )
        equator = self._equator.eval(data_array, column, )
        jacobian = self._jacobian.eval(data_array, column, )
        self.iter_printer.next(maybelog_guess, equator, jacobian_calculated=True, )
        return equator, jacobian

    #]




"""
Dynamic nonlinear period-by-period simulator
"""


#[
from __future__ import annotations

import numpy as _np
import scipy as _sp
import functools as _ft
import wlogging as _wl

from ..evaluators.base import (DEFAULT_INIT_GUESS, )
from ..simultaneous import main as _simultaneous
from ..plans.simulation_plans import (SimulationPlan, )
from ..dataslates.main import (Dataslate, )
from .. import equations as _equations
from .. import wrongdoings as _wrongdoings
from .evaluators import (PeriodEvaluator, )
from ..fords import simulators as _ford_simulators

from typing import (TYPE_CHECKING, )
if TYPE_CHECKING:
    from numbers import (Real, )
    from typing import (Callable, Any, Literal, )
    from ..dates import (Period, )
#]


_DEFAULT_FALLBACK_VALUE = 1/9


def simulate(
    simulatable_v: _simultaneous.Simultaneous,
    dataslate_v: Dataslate,
    plan: SimulationPlan | None,
    vid: int,
    *,
    logger: _wl.Logger,
    start_iter_from: Literal["data", "first_order", "previous_period"] = "previous_period",
    root_settings: dict[str, Any] | None = None,
    iter_printer_settings: dict[str, Any] | None = None,
    when_fails: Literal["critical", "error", "warning", "silent"] = "critical",
    when_missing: Literal["critical", "error", "warning", "silent"] = "critical",
    fallback_value: Real = _DEFAULT_FALLBACK_VALUE,
    terminal: Literal["data", "first_order", ] = "first_order",
    return_info: bool = False,
) -> dict[str, Any]:
    """
    """
    if not simulatable_v.is_singleton:
        raise ValueError("Simulator requires a singleton simulatable object")
    #
    root_settings = (
        _DEFAULT_ROOT_SETTINGS
        if root_settings is None
        else _DEFAULT_ROOT_SETTINGS | root_settings
    )

    qid_to_name = simulatable_v.create_qid_to_name()
    name_to_qid = simulatable_v.create_name_to_qid()
    all_quantities = simulatable_v.get_quantities()
    #
    wrt_qids = _get_wrt_qids(simulatable_v, name_to_qid, )
    wrt_equations = simulatable_v.get_dynamic_equations(kind=_equations.ENDOGENOUS_EQUATION, )

    if len(wrt_equations) != len(wrt_qids):
        raise _wrongdoings.IrisPieError(
            f"Number of endogenous equations {len(wrt_equations)} "
            f"does not match number of endogenous quantities {len(wrt_qids)}"
        )

    if not wrt_equations:
        return

    terminal_simulator = (
        _ford_simulators.create_terminal_simulator(simulatable_v, )
        if terminal == "first_order" else None
    )

    evaluator_factory = _ft.partial(
        PeriodEvaluator,
        wrt_equations=wrt_equations,
        all_quantities=all_quantities,
        terminal_simulator=terminal_simulator,
        context=simulatable_v.get_context(),
        iter_printer_settings=iter_printer_settings,
    )

    base_evaluator = evaluator_factory(wrt_qids, )

    when_missing_stream = \
        _wrongdoings.STREAM_FACTORY[when_missing] \
        (f"These values are missing at the start of simulation: ")

    starter = _ITER_STARTER[start_iter_from]

    catch_missing = _ft.partial(
        _catch_missing,
        dataslate_v=dataslate_v,
        qid_to_name=qid_to_name,
        fallback_value=fallback_value,
        when_missing_stream=when_missing_stream,
    )
    #
    periods = dataslate_v.periods
    data = dataslate_v.get_data_variant(0, )
    #
    for t in dataslate_v.base_columns:
        current_wrt_qids, current_evaluator = \
            _set_up_current_period(plan, evaluator_factory, wrt_qids, periods[t], base_evaluator, name_to_qid, )
        current_evaluator.iter_printer.header_message = \
            _create_header_message(vid, t, periods[t], )
        data[current_wrt_qids, t] = starter(data, current_wrt_qids, t, )
        catch_missing(data, t, )
        #
        init = current_evaluator.get_init_guess(data, t, )
        root_final = _sp.optimize.root(
            current_evaluator.eval, init,
            args=(data, t, ),
            jac=True,
            **root_settings,
        )
        #
        if not root_final.success:
            _catch_fail(root_final, when_fails, )
        current_evaluator.update(root_final.x, data, t, )
        #
        current_evaluator.iter_printer.print_footer()
        current_evaluator.iter_printer.reset()
    #
    info = {
        "method": "period",
    }
    #
    return info


# REFACTOR
def _set_up_current_period(
    plan: SimulationPlan | None,
    evaluator_factory: Callable[..., PeriodEvaluator],
    base_wrt_qids: tuple[int, ...],
    current_period: Period,
    base_evaluator: PeriodEvaluator,
    name_to_qid: dict[str, int],
    /,
) -> tuple[tuple[int, ...], PeriodEvaluator]:
    """
    """
    if plan is None:
        return base_wrt_qids, base_evaluator
    #
    names_exogenized = plan.get_exogenized_unanticipated_in_period(current_period, )
    names_endogenized = plan.get_endogenized_unanticipated_in_period(current_period, )
    if not names_exogenized and not names_endogenized:
        return base_wrt_qids, base_evaluator
    #
    if len(names_exogenized) != len(names_endogenized):
        raise _wrongdoings.IrisPieCritical(
            f"Number of exogenized quantities {len(names_exogenized)}"
            f" does not match number of endogenized quantities {len(names_endogenized)}"
            f" in period {current_period}"
        )
    #
    qids_exogenized = tuple(name_to_qid[name] for name in names_exogenized)
    qids_endogenized = tuple(name_to_qid[name] for name in names_endogenized)
    current_wrt_qids = tuple(sorted(set(base_wrt_qids).difference(qids_exogenized).union(qids_endogenized)))
    current_evaluator = evaluator_factory(current_wrt_qids, )
    return current_wrt_qids, current_evaluator


def _start_iter_from_previous_period(
    data: _np.ndarray,
    wrt_qids: tuple[int, ...],
    t: int,
) -> _np.ndarray:
    """
    """
    source_t = t - 1 if t > 0 else 0
    return data[wrt_qids, source_t]


def _start_iter_from_data(
    data: _np.ndarray,
    wrt_qids: tuple[int, ...],
    t: int,
) -> _np.ndarray:
    """
    """
    return data[wrt_qids, t]


def _start_iter_from_first_order(
    data: _np.ndarray,
    wrt_qid: tuple[int, ...],
    t: int,
) -> _np.ndarray:
    """
    """
    raise NotImplementedError()


def _catch_missing(
    data: _np.ndarray,
    t: int,
    /,
    dataslate_v: Dataslate,
    qid_to_name: dict[int, str],
    fallback_value: Real,
    when_missing_stream: _wrongdoings.Stream,
) -> None:
    """
    """
    missing = _np.isnan(data[:, t])
    if not missing.any():
        return
    #
    data[missing, t] = fallback_value
    #
    current_period = dataslate_v.periods[t]
    shift = 0
    for qid in _np.flatnonzero(missing):
        when_missing_stream.add(
            f"{qid_to_name[qid]}[{current_period+shift}]"
            f" when simulating {current_period}"
        )


def _catch_fail(
    root_final: _sp.optimize.OptimizeResult,
    when_fails: Literal["error", "warning", "silent"],
) -> None:
    """
    """
    raise ValueError()


def _create_header_message(
    vid: str,
    t: int,
    current_period: int,
) -> str:
    """
    """
    return f"[Variant {vid}][Period {current_period}]"


def _get_wrt_qids(
    simulatable_v,
    name_to_qid: dict[str, int],
    /,
) -> tuple[int, ...]:
    """
    """
    plannable = simulatable_v.get_simulation_plannable()
    wrt_names \
        = tuple(plannable.can_be_exogenized_unanticipated) \
        + tuple(plannable.can_be_exogenized_anticipated)
    wrt_qids = sorted(
        name_to_qid[name]
        for name in set(wrt_names)
    )
    return wrt_qids


_ITER_STARTER = {
    "previous_period": _start_iter_from_previous_period,
    "data": _start_iter_from_data,
    "first_order": _start_iter_from_first_order,
}


_DEFAULT_ROOT_SETTINGS = {
    "method": "hybr",
    "tol": 1e-12,
}





import irispie as ir


def test_regular_spans_shifted():
    x = ir.Series(periods=ir.qq(2020,1)>>ir.qq(2025,4), values=1, )
    y = ir.Series(periods=ir.qq(2021,4)>>ir.qq(2026,4), values=1, )
    (span, start, end) = ir.dates.get_encompassing_span(x, y, )
    assert span.start == x.start
    assert start == x.start
    assert span.end == y.end
    assert end == y.end


def test_regular_spans_nested():
    x = ir.Series(periods=ir.qq(2020,1)>>ir.qq(2026,4), values=1, )
    y = ir.Series(periods=ir.qq(2021,4)>>ir.qq(2025,4), values=1, )
    (span, start, end) = ir.dates.get_encompassing_span(x, y, )
    assert span.start == x.start
    assert start == x.start
    assert span.end == x.end
    assert end == x.end


def test_one_empty_series():
    x = ir.Series(periods=ir.qq(2020,1)>>ir.qq(2025,4), values=1, )
    y = ir.Series()
    (span, start, end) = ir.dates.get_encompassing_span(x, y, )
    assert span.start == x.start
    assert start == x.start
    assert span.end == x.end
    assert end == x.end


def test_both_empty_series():
    x = ir.Series()
    y = ir.Series()
    (span, start, end) = ir.dates.get_encompassing_span(x, y, )
    assert span.start is ir.start
    assert start is None
    assert span.end is ir.end
    assert end is None





"""
"""


#[
from __future__ import annotations

from typing import (TYPE_CHECKING, )
import warnings as _wa
import copy as _co
import numpy as _np
import operator as _op

from ..conveniences import copies as _copies
from .. import quantities as _quantities

if TYPE_CHECKING:
    from numbers import (Real, )
    from typing import (Self, Literal, Callable, )
    from collections.abc import (Iterable, )
#]


class Variant:
    """
    Container for parameter variant specific attributes of a model
    """
    #[

    __slots__ = (
        "levels",
        "changes",
        "solution",
        "_max_qid",
    )

    def __init__(self, ) -> None:
        """
        """
        self.levels = None
        self.changes = None
        self.solution = None
        self._max_qid = None

    @classmethod
    def from_source(
        klass,
        quantities: Iterable[_quantities.Quantity],
        is_flat: bool,
        /,
        **kwargs,
    ) -> Self:
        """
        """
        self = klass()
        max_qid = _quantities.get_max_qid(quantities, )
        self._max_qid = max_qid
        self.solution = None
        self._initilize_values()
        if is_flat:
            qid_to_logly = _quantities.create_qid_to_logly(quantities, )
            self.zero_changes(qid_to_logly, )
        return self

    def copy(self, /, ) -> Self:
        """
        """
        new = type(self)()
        for i in ("levels", "changes", "solution", ):
            attr = getattr(self, i, )
            if attr is not None:
                setattr(new, i, attr.copy(), )
        new._max_qid = self._max_qid
        return new

    def _initilize_values(self, /, ) -> None:
        """
        """
        self._reset_levels()
        self._reset_changes()

    def _reset_levels(self, /, ) -> None:
        """
        """
        self.levels = _np.full((self._max_qid+1, ), _np.nan, dtype=float, )

    def _reset_changes(self, /, ) -> None:
        """
        """
        self.changes = _np.full((self._max_qid+1, ), _np.nan, dtype=float, )

    def update_values_from_dict(self, update: dict, ) -> None:
        self.levels = _update_from_dict(self.levels, update, _op.itemgetter(0), lambda x: x, )
        self.changes = _update_from_dict(self.changes, update, _op.itemgetter(1), lambda x: ..., )

    def update_levels_from_array(self, levels: _np.ndarray, qids: Iterable[int], ) -> None:
        _update_from_array(self.levels, levels, qids, )

    def update_changes_from_array(self, changes: _np.ndarray, qids: Iterable[int], ) -> None:
        _update_from_array(self.changes, changes, qids, )

    def retrieve_values(
        self,
        attr: Literal["levels", "changes"],
        qids: Iterable[int] | None = None,
        /,
    ) -> _np.ndarray:
        values = _np.copy(getattr(self, attr, ), )
        return values[qids, ...]

    def rescale_values(
        self,
        attr: Literal["levels", "changes"],
        factor: Real,
        qids: Iterable[int],
    ) -> None:
        """
        Rescale values by a common factor
        """
        attr = getattr(self, attr, )
        qids = list(qids) if qids is not None else None
        attr[qids, ...] *= factor

    def retrieve_maybelog_values_for_qids(
        self,
        qids: Iterable[int],
        qid_to_logly: dict[int, bool],
    ) -> tuple[_np.ndarray, ..., ]:
        """
        """
        qids = list(qids)
        where_logly = list(_quantities.generate_where_logly(qids, qid_to_logly, ))
        #
        # Extract initial guesses for levels and changes
        maybelog_levels = self.levels[qids].flatten()
        maybelog_changes = self.changes[qids].flatten()
        #
        # Logarithmize
        maybelog_levels[where_logly] = _np.log(maybelog_levels[where_logly])
        maybelog_changes[where_logly] = _np.log(maybelog_changes[where_logly])
        #
        return maybelog_levels, maybelog_changes

    def zero_changes(
        self,
        qid_to_logly: dict[int, bool | None],
        /,
    ) -> None:
        """
        Reset all quantities to flat
        """
        self._reset_changes()
        for qid, logly in qid_to_logly.items():
            if logly is None:
                continue
            self.changes[qid] = 1 if logly else 0

    def create_steady_array(
        self,
        qid_to_logly: dict[int, bool],
        /,
        num_columns: int = 1,
        shift_in_first_column: int = 0,
    ) -> _np.ndarray:
        """
        """
        levels = _np.copy(self.levels).reshape(-1, 1)
        changes = _np.copy(self.changes).reshape(-1, 1)
        #
        if num_columns==1 and shift_in_first_column==0:
            return levels
        #
        qids = list(range(self._max_qid+1))
        where_logly = list(_quantities.generate_where_logly(qids, qid_to_logly))
        #
        shift_vec = _np.array(range(shift_in_first_column, shift_in_first_column+num_columns))
        #
        _wa.filterwarnings(action="ignore", category=RuntimeWarning)
        levels[where_logly] = _np.log(levels[where_logly])
        changes[where_logly] = _np.log(changes[where_logly])
        _wa.filterwarnings(action="default", category=RuntimeWarning)
        #
        levels[_np.isnan(levels) | _np.isinf(levels)] = _np.nan
        changes[_np.isnan(changes) | _np.isinf(changes)] = 0
        #
        steady_array = levels + changes * shift_vec
        #
        _wa.filterwarnings(action="ignore", category=RuntimeWarning)
        steady_array[where_logly, :] = _np.exp(steady_array[where_logly, :])
        _wa.filterwarnings(action="default", category=RuntimeWarning)
        #
        return steady_array

    def create_zero_array(
        self,
        qid_to_logly: dict[int, bool],
        /,
        num_columns: int = 1,
        shift_in_first_column: int = 0,
    ) -> _np.ndarray:
        """
        """
        levels = _np.copy(self.levels).reshape(-1, 1)
        inx_set_to_0 = [ qid for qid, logly in qid_to_logly.items() if logly is False ]
        inx_set_to_1 = [ qid for qid, logly in qid_to_logly.items() if logly is True ]
        levels[inx_set_to_0] = 0
        levels[inx_set_to_1] = 1
        return _np.tile(levels, (1, num_columns))

    #]


def _update_from_array(
    values: _np.ndarray,
    updated_values: _np.ndarray | Iterable[Real] | None,
    qids: Iterable[int],
) -> None:
    """
    Update levels or changes from an array and a list of qids
    """
    #[
    if updated_values is None:
        return
    if hasattr(updated_values, "flat"):
        values[list(qids)] = updated_values.flat
        return
    values[list(qids)] = updated_values
    #]


def _update_from_dict(
    what_to_update: _np.ndarray,
    update: dict[int, Real | tuple[Real]],
    when_tuple: Callable,
    when_not_tuple: Callable,
    /,
) -> _np.ndarray:
    """
    Update levels or changes from a dictionary
    """
    #[
    for qid, value in update.items():
        new_value = when_tuple(value) if isinstance(value, tuple) else when_not_tuple(value)
        what_to_update[qid] = new_value if new_value is not ... else what_to_update[qid]
    return what_to_update
    #]





import pytest
import random as rn
import numpy as np
import irispie as ir


rn.seed(0)

N = 100

def random_period():
    freq = rn.choices([1, 2, 4, 12], )[0]
    year = rn.randint(2000, 2050, )
    per = rn.randint(1, freq, )
    return ir.Period.from_year_period(freq, year, per, )


def random_series():
    return ir.Series(
        start=random_period(),
        values=rn.gauss(),
        num_variants=rn.randint(1, 5, ),
    )


@pytest.mark.parametrize(
    ["x", "horizon", ], [(random_series(), rn.randint(4,10)) for _ in range(N, )],
)
def test_extrapolate(x: ir.Series, horizon: int, ) -> None:
    y = ir.extrapolate(x, (0.8, ), x.end+1 >> x.end+horizon, intercept=0, )
    x_end = x[x.end, :].flatten()
    y_end = y[y.end, :].flatten()
    assert np.allclose(x_end * 0.8**horizon, y_end, atol=1e-6, rtol=1e-6, )




"""
Jacobians for dynamic period-by-period systems
"""


#[
from __future__ import annotations

import numpy as _np
from types import (SimpleNamespace, )

from ..incidences.main import (Token, )
from ..incidences import main as _incidence
from ..aldi.maps import (ArrayMap, )
from ..equations import (Equation, )
from ..jacobians import base

from typing import (TYPE_CHECKING, )
if TYPE_CHECKING:
    from typing import (Any, )
    from collections.abc import (Collection, Iterable, )
    from ..aldi.differentiators import (AtomFactoryProtocol, )
#]


# Implement AtomFactoryProtocol


def _flat_create_diff_for_token(
    token: Token,
    wrt_qids: tuple[int],
    /,
) -> _np.ndarray | int:
    """
    """
    try:
        index = wrt_qids.index(token.qid, )
    except ValueError:
        return 0
    diff = _np.zeros((len(wrt_qids), 1, ))
    diff[index, 0] = 1
    return diff


_FLAT_ATOM_FACTORY = SimpleNamespace(
    create_data_index_for_token=lambda token: (token.qid, token.shift, ),
    create_diff_for_token=_flat_create_diff_for_token,
    get_diff_shape=lambda wrt_qids: (len(wrt_qids), 1, ),
)


def _nonflat_create_diff_for_token(
    token: Token,
    wrt_qids: tuple[int],
    /,
) -> _np.ndarray | int:
    """
    """
    try:
        index = wrt_qids.index(token.qid, )
    except ValueError:
        return 0
    diff = _np.zeros((len(wrt_qids), 2, ))
    diff[index, :] = 1, token.shift,
    return diff


_NONFLAT_ATOM_FACTORY = SimpleNamespace(
    create_data_index_for_token=lambda token: (token.qid, token.shift, ),
    create_diff_for_token=_nonflat_create_diff_for_token,
    get_diff_shape=lambda wrt_qids: (len(wrt_qids), 2, ),
)



class _SteadyJacobian(base.Jacobian, ):
    """
    """
    #[

    def _populate_map(self, *args, **kwargs, ) -> None:
        """
        """
        self._map = ArrayMap.static(*args, **kwargs, )

    def eval(*args, **kwargs, ): raise NotImplementedError

    def _create_eid_to_wrts(
        self,
        equations: Iterable[Equation],
        all_wrt_qids: Iterable[int],
        /,
    ) -> dict[int, tuple[int, ...]]:
        """
        Create {eid: (qid, ... ), ...} where (qid, ... ) are qids wrt which the
        equation is defined no matter the time shift of the token
        """
        return {
            eqn.id: tuple(
                qid for qid in all_wrt_qids
                if _incidence.is_qid_in_tokens(eqn.incidence, qid, )
            )
            for eqn in equations
        }

        #]


class FlatSteadyJacobian(_SteadyJacobian, ):
    """
    """
    #[

    _atom_factory = _FLAT_ATOM_FACTORY

    def eval(
        self,
        steady_array: _np.ndarray,
        column_offset: int | None,
        /,
    ) -> _np.ndarray:
        """
        """
        diff_array = self._aldi_context.eval_diff_to_array(steady_array, column_offset, )
        return self._create_jacobian_matrix(diff_array, )

    #]


class NonflatSteadyJacobian(_SteadyJacobian, ):
    """
    """
    #[

    # Assigned in the evaluator
    NONFLAT_STEADY_SHIFT = ...

    _atom_factory = _NONFLAT_ATOM_FACTORY

    def eval(
        self,
        steady_array: _np.ndarray,
        column_offset: int,
        /,
    ) -> _np.ndarray:
        """
        """
        diff_array = self._aldi_context.eval_diff_to_array(steady_array, column_offset, )
        A = self._create_jacobian_matrix(diff_array[:, 0:1], )
        B = self._create_jacobian_matrix(diff_array[:, 1:2], )
        k = self.NONFLAT_STEADY_SHIFT
        return _np.vstack((
            _np.hstack((A, B, )),
            _np.hstack((A, B+k*A, )),
        ))

    #]




"""
"""


#[
from __future__ import annotations
#]


class HasInvariantMixin:
    """
    """
    #[

    def get_description(self, /, ) -> str:
        """
        """
        return self._invariant.get_description()

    def set_description(self, *args, **kwargs, ) -> None:
        """
        """
        self._invariant.set_description(*args, **kwargs, )

    #]




"""
Meta plans
"""


from .simulation_plans import *
from .simulation_plans import __all__ as simulation_plans_all

from .steady_plans import *
from .steady_plans import __all__ as steady_plans_all

__all__ = (
    *simulation_plans_all,
    *steady_plans_all,
)




"""
Jacobians for dynamic period-by-period systems
"""


#[
from __future__ import annotations

from types import (SimpleNamespace, )
import numpy as _np

from ..jacobians import base
from ..aldi.maps import (ArrayMap, )

from typing import (TYPE_CHECKING, )
if TYPE_CHECKING:
    from types import (EllipsisType, )
    from collections.abc import (Iterable, )
    from ..incidences.main import (Token, )
    from ..equations import (Equation, )
#]


class Jacobian(base.Jacobian, ):
    """
    """
    #[

    def _populate_map(
        self,
        eids: list[int],
        eid_to_wrt_tokens: dict[int, Iterable[Token]],
        tokens_in_columns_on_lhs: list[Token],
        eid_to_rhs_offset: dict[int, int],
        **kwargs,
    ) -> None:
        """
        """
        map = ArrayMap()
        token_to_lhs_column = {
            t: i
            for i, t in enumerate(tokens_in_columns_on_lhs, )
        }
        num_eids = len(eids)
        #
        map_tuples = []
        rhs_row_offset = 0
        for eqn_enum, eid in enumerate(eids, ):
            wrt_tokens = eid_to_wrt_tokens[eid]
            for rhs_row, tok in enumerate(wrt_tokens, start=rhs_row_offset, ):
                for rhs_column in self._columns_to_eval:
                    try:
                        tok_shifted = tok.shifted(rhs_column, )
                        lhs_column = token_to_lhs_column[tok_shifted]
                        lhs_row = eqn_enum + num_eids*rhs_column
                        map_tuples.append((lhs_row, lhs_column, rhs_row, rhs_column, ))
                    except KeyError:
                        continue
            rhs_row_offset += len(wrt_tokens)
        #
        lhs_rows, lhs_columns, rhs_rows, rhs_columns = zip(*map_tuples, )
        map.lhs = (list(lhs_rows), list(lhs_columns), )
        # map.lhs = (list(lhs_columns), list(lhs_rows), )
        map.rhs = (list(rhs_rows), list(rhs_columns), )
        self._map = map

    def eval(
        self,
        data_array: _np.ndarray,
        column_offset: _np.ndarray,
    ) -> _np.ndarray:
        """
        """
        diff_array = self._aldi_context.eval_diff_to_array(data_array, column_offset, )
        extended_jacobian_matrix = self._create_jacobian_matrix(diff_array, )
        jacobian_matrix = extended_jacobian_matrix[:, :42]
        for i in range(8):
            jacobian_matrix += 0.*extended_jacobian_matrix[:, 42+i*42:42+i*42+42]
        # j = jacobian_matrix
        # x = extended_jacobian_matrix
        # import IPython; IPython.embed(header=__name__)
        return jacobian_matrix

    def _create_eid_to_wrts(
        self,
        equations: Iterable[Equation],
        all_wrt_tokens: Iterable[int],
        /,
    ) -> dict[int, tuple[int, ...]]:
        """
        """
        all_wrt_qids = set(tok.qid for tok in all_wrt_tokens)
        return {
            eqn.id: tuple(
                tok for tok in eqn.incidence
                if tok.qid in all_wrt_qids
            )
            for eqn in equations
        }

    @property
    def _atom_factory(self, /, ) -> AtomFactoryProtocol:
        """
        """
        columns_to_eval = self._columns_to_eval
        num_columns_to_eval = self._num_columns_to_eval
        #
        def create_diff_for_token(
            self,
            token: Token,
            wrt_tokens: tuple[Token],
        ) -> _np.ndarray | int:
            """
            """
            num_wrts = len(wrt_tokens)
            try:
                index = wrt_tokens.index(token, )
            except ValueError:
                return 0
            diff = _np.zeros((num_wrts, self._num_columns_to_eval, ), dtype=_np.float64, )
            diff[index] = 1
            return diff
        #
        return SimpleNamespace(
            create_data_index_for_token=lambda token: (token.qid, token.shift + _np.array(columns_to_eval, ), ),
            create_diff_for_token=create_diff_for_token,
            get_diff_shape=lambda wrt_tokens: (len(wrt_tokens), num_columns_to_eval, ),
        )

    #]




"""
"""


#[
from __future__ import annotations

import numpy as _np
import scipy as _sp
#]


_ELEMENTWISE_FUNCTIONS = {
    "log": _np.log,
    "exp": _np.exp,
    "sqrt": _np.sqrt,
    "logistic": _sp.special.expit,
    "maximum": _np.maximum,
    "minimum": _np.minimum,
}


for n in _ELEMENTWISE_FUNCTIONS.keys():
    exec(
        f"def {n}(x, *args, **kwargs):\n"
        f"    if hasattr(x, '{n}'):\n"
        f"        return x.{n}(*args, **kwargs, )\n"
        f"    else:\n"
        f"        return _ELEMENTWISE_FUNCTIONS['{n}'](x, *args, **kwargs, )\n"
    )


def add_function_adaptations_to_context(context: dict | None) -> dict:
    """
    """
    #[
    context = context if context else {}
    for n in _ELEMENTWISE_FUNCTIONS.keys():
        context[n] = globals()[n]
    return context
    #]




"""
"""


#[
from __future__ import annotations

from typing import (Iterable, )
import requests as _rq

from .. import dates as _dates
from ..dates import (Frequency, )
from ..series.main import (Series, )

from . import main as _databoxes
#]


_FRED_FREQ_MAP = {
    "A".casefold(): Frequency.YEARLY,
    "Q".casefold(): Frequency.QUARTERLY,
    "M".casefold(): Frequency.MONTHLY,
    "D".casefold(): Frequency.DAILY,
}


_BASE_URL = r"https://api.stlouisfed.org/fred/series"
_OBSERVATIONS_URL = _BASE_URL + r"/observations"
_API_KEY = r"951f01181da86ccb9045ce8716f82f43"
_PARAMETERS = r"?series_id={series_id}&api_key={api_key}&file_type=json"
_MISSING_VALUE = r"."


class FredMixin:
    """
    """
    #[

    @classmethod
    def from_fred(
        klass,
        mapper: Iterable[str] | dict[str, str],
        /,
    ) -> _databoxes.Databox:
        """
        """
        self = klass()
        if not isinstance(mapper, dict):
            mapper = _mapper_from_series_ids(mapper, )
        for name, series_id in mapper.items():
            self[name] = _get_series(series_id, )
        return self

    #]


def _mapper_from_series_ids(series_ids: Iterable[str], /, ):
    """
    """
    #[
    return {
        series_id.strip(): series_id.strip()
        for series_id in series_ids
    }
    #]

def _get_series(series_id: str, /, ):
    """
    """
    #[
    urls = _get_series_urls(series_id, )
    meta_response = _rq.get(urls["meta_url"], ).json()
    data_response = _rq.get(urls["data_url"], ).json()
    #
    if "seriess" not in meta_response or "observations" not in data_response:
        raise IriePieCritical(f"Invalid response from FRED API for {series_id}", )
    #
    freq = _get_freq_from_meta_response(meta_response, )
    iso_dates, str_values = _get_dates_and_values_from_data_response(data_response, )
    values = ( (float(x) if x != _MISSING_VALUE else None) for x in str_values )
    dates = _dates.periods_from_iso_strings(iso_dates, frequency=freq, )
    return Series(dates=dates, values=list(values), )
    #]


def _get_freq_from_meta_response(meta_response: dict, /, ):
    """
    """
    #[
    freq_letter = meta_response["seriess"][0]["frequency_short"].casefold()
    return _FRED_FREQ_MAP[freq_letter]
    #]


def _get_dates_and_values_from_data_response(data_response: dict, /, ):
    """
    """
    #[
    date_value_pairs = (
        (obs["date"], obs["value"], )
        for obs in data_response["observations"]
    )
    return zip(*date_value_pairs, )
    #]


def _get_series_urls(series_id: str, /, ):
    """
    """
    #[
    parameters = _PARAMETERS.format(series_id=series_id, api_key=_API_KEY, )
    meta_url = _BASE_URL + parameters
    data_url = _OBSERVATIONS_URL + parameters
    return {"meta_url": meta_url, "data_url": data_url, }
    #]




"""
First-order terminators
========================

"""


#[
from __future__ import annotations

from types import (SimpleNamespace, )
import numpy as _np

from . import simulators as _simulators

from typing import (TYPE_CHECKING, )
if TYPE_CHECKING:
    from typing import (Callable, )
    from .simulators import FordSimulatableProtocol
#]


def create_terminators(
    simulatable_v: FordSimulatableProtocol,
    terminal: Literal["data", "first_order", ],
) -> SimpleNamespace:
    """
    """
    #[
    max_lead = simulatable_v.max_lead
    if terminal != "first_order" or not max_lead:
        return SimpleNamespace(
            terminate_simulation=None,
            terminate_jacobian=None,
        )
    #
    import ipdb; ipdb.set_trace()
    solution = simulatable_v.get_singleton_solution(deviation=False, )
    vec = simulatable_v.solution_vectors
    curr_xi_qids, curr_xi_indexes = vec.get_curr_transition_indexes()
    T = solution.T
    K = solution.K
    #
    # curr_TT := [ T; T @ T; ... ] with current dated rows only
    # curr_KK := [ K; T @ K + K; ... ] with current dated rows only
    curr_TT = []
    curr_KK = []
    cum_T = _np.eye(*T.shape, dtype=T.dtype, )
    cum_K = _np.zeros(K.shape, dtype=K.dtype, )
    for i in range(max_lead, ):
        cum_T = T @ cum_T
        cum_K = T @ cum_K + K
        curr_TT.append(cum_T[curr_xi_indexes, ...])
        curr_KK.append(cum_K[curr_xi_indexes, ...])
    curr_TT = _np.vstack(curr_TT, )
    curr_KK = _np.hstack(curr_KK, )
    #
    qid_to_logly = simulatable_v.create_qid_to_logly()
    logly_rows = tuple( qid for qid, status in qid_to_logly.items() if status )
    #
    #
    def terminate_simulation(
        data_array: _np.ndarray,
        last_simulation: int,
    ) -> None:
        """
        """
        first_terminal = last_simulation + 1
        terminal_columns = tuple(range(first_terminal, first_terminal + max_lead, ), )
        data_array[logly_rows, :] = _np.log(data_array[logly_rows, :])
        terminit_xi = _simulators.get_init_xi(data_array, first_terminal, vec, )
        #
        terminal_curr_xi = curr_TT @ terminit_xi + curr_KK
        #
        terminal_curr_xi = terminal_curr_xi.reshape(-1, max_lead, order="F", )
        data_array[_np.ix_(curr_xi_qids, terminal_columns)] = terminal_curr_xi
        data_array[logly_rows, :] = _np.exp(data_array[logly_rows, :])
    #
    #
    def terminate_jacobian(
    ) -> None:
        """
        """
        pass
    #
    #
    return SimpleNamespace(
        terminate_simulation=terminate_simulation,
        terminate_jacobian=terminate_jacobian,
    )
    #]





"""
Jacobians for dynamic period-by-period systems
"""


#[
from __future__ import annotations

from types import (SimpleNamespace, )
from collections.abc import (Iterable, )
import numpy as _np

from ..incidences.main import (Token, )
from ..incidences import main as _incidence
from ..aldi.maps import (ArrayMap, )
from ..aldi.differentiators import (AtomFactoryProtocol, )

from ..jacobians import base
#]


# Implement AtomFactoryProtocol

def _create_diff_for_token(
    token: Token,
    wrt_qids: tuple[int],
    /,
) -> _np.ndarray | int:
    """
    """
    if token.shift != 0:
        return 0
    try:
        index = wrt_qids.index(token.qid, )
    except ValueError:
        return 0
    diff = _np.zeros((len(wrt_qids), 1, ), )
    diff[index] = 1
    return diff


_ATOM_FACTORY = SimpleNamespace(
    create_data_index_for_token=lambda token: (token.qid, token.shift, ),
    create_diff_for_token=_create_diff_for_token,
    get_diff_shape=lambda wrt_qids: (len(wrt_qids), 1, ),
)


class Jacobian(base.Jacobian, ):
    """
    """
    #[

    _atom_factory: AtomFactoryProtocol = _ATOM_FACTORY

    def _populate_map(self, *args, **kwargs, ) -> None:
        """
        """
        self._map = ArrayMap.static(*args, **kwargs, )

    def eval(
        self,
        data_array: _np.ndarray,
        column_offset: int,
        /,
    ) -> _np.ndarray:
        """
        """
        diff_array = self._aldi_context.eval_diff_to_array(data_array, column_offset, )
        return self._create_jacobian_matrix(diff_array, )

    def _create_eid_to_wrts(
        self,
        equations: Iterable[_equations.Equation],
        all_wrt_qids: Iterable[Tokens],
        /,
    ) -> dict[int, tuple[Token, ...]]:
        """
        """
        return {
            eqn.id: tuple(
                qid for qid in all_wrt_qids
                if _incidence.is_qid_zero_in_tokens(eqn.incidence, qid, )
            )
            for eqn in equations
        }

    #]





"""
"""

#[
from __future__ import annotations

from collections.abc import (Iterable, )
from typing import (Any, )

from ..equations import (Equation, )
from ..quantities import (Quantity, )
from .. import quantities as _quantities
from ..explanatories import main as _explanatories
from ..conveniences import descriptions as _descriptions
#]


class Invariant(
    _descriptions.DescriptionMixin,
):
    """
    """
    #[

    __slots__ = (
        "explanatories",
        "lhs_names",
        "residual_names",
        "rhs_only_names",
        "parameter_names",
        "_context",
        "_description",
    )

    def __init__(self, /, ) -> None:
        """
        """
        self.explanatories = ()
        self.lhs_names = ()
        self.residual_names = ()
        self.rhs_only_names = ()
        self.parameter_names = ()
        self._context = {}
        self._description = ""

    @classmethod
    def from_equations(
        klass,
        equations: Iterable[Equation],
        quantities: Iterable[Quantity] | None,
        /,
        context: dict[str, Any] | None = None,
        description: str | None = None,
        **kwargs,
    ) -> None:
        """
        """
        self = klass()
        self._context = context or {}
        self.set_description(description, )
        self.explanatories = tuple(
            _explanatories.Explanatory(e, context=self._context, **kwargs, )
            for e in equations
        )
        self.collect_names()
        self.finalize_explanatories()
        quantity_names = _quantities.generate_all_quantity_names(quantities, )
        self.parameter_names = tuple(n for n in quantity_names if n in self.all_names)
        return self

    @property
    def num_equations(self, /, ) -> int:
        """
        Number of equations.
        """
        return len(self.explanatories)

    @property
    def num_lhs_names(self, /, ) -> int:
        """
        Number of unique LHS names
        """
        return len(self.lhs_names)

    @property
    def equations(self, /, ) -> tuple[Equation]:
        return tuple( x.equation for x in self.explanatories )

    def collect_names(self, /, ) -> None:
        """
        """
        self.collect_lhs_names()
        self.collect_residual_names()
        self.collect_rhs_only_names()

    def collect_lhs_names(self, /, ) -> None:
        """
        Tuple of names of LHS variables in order of their first appearance in the equations
        """
        self.lhs_names = []
        for x in self.explanatories:
            if x.lhs_name not in self.lhs_names:
                self.lhs_names.append(x.lhs_name)
        self.lhs_names = tuple(self.lhs_names)

    def collect_residual_names(self, /, ) -> tuple[str]:
        """
        Tuple of names of LHS variables in order of appearance
        """
        self.residual_names = []
        for x in self.explanatories:
            if x.residual_name is None:
                continue
            if x.residual_name in self.residual_names:
                continue
            self.residual_names.append(x.residual_name, )
        self.residual_names = tuple(self.residual_names, )

    def collect_rhs_only_names(
        self,
        /,
    ) -> None:
        """
        """
        all_names = set()
        for x in self.explanatories:
            all_names.update(x.all_names, )
        self.rhs_only_names = tuple(all_names.difference(self.lhs_names + self.residual_names, ))

    @property
    def all_names(self, /, ) -> tuple[str]:
        """
        """
        return self.lhs_names + self.rhs_only_names + self.residual_names

    def finalize_explanatories(
        self,
        /,
    ) -> None:
        """
        """
        name_to_qid = self.create_name_to_qid()
        for x in self.explanatories:
            x.finalize(name_to_qid, )

    def reorder_equations(
        self,
        new_order = Iterable[int],
        /,
    ) -> None:
        """
        """
        # Check if new_order is a valid permutation
        if sorted(new_order) != list(range(self.num_equations)):
            raise ValueError("New equation order must be a permutation of integers from 0 to num_equations-1")
        #
        self.explanatories = [
            self.explanatories[i]
            for i in new_order
        ]
        # We need to recollect the names to make sure the LHS names and the
        # residual names are in the right order consistent with the new order
        # of the equations
        self.collect_names()
        self.finalize_explanatories()

    def create_name_to_qid(self, /, ) -> dict[str, int]:
        """
        """
        return { name: i for i, name in enumerate(self.all_names) }

    def create_name_to_qid(self, /, ) -> dict[str, int]:
        """
        """
        return { name: i for i, name in enumerate(self.all_names) }

    #]

